@@ -1,41 +1,28 @@
-var N3Lexer = require('../N3').Lexer;
+import { Lexer } from '../src/';
+import queueMicrotask from 'queue-microtask';
 
-var EventEmitter = require('events');
+import { EventEmitter } from 'events';
 
-describe('N3Lexer', function () {
-  describe('The N3Lexer module', function () {
-    it('should be a function', function () {
-      N3Lexer.should.be.a('function');
+describe('Lexer', () => {
+  describe('The Lexer export', () => {
+    it('should be a function', () => {
+      Lexer.should.be.a('function');
     });
 
-    it('should make N3Lexer objects', function () {
-      N3Lexer().should.be.an.instanceof(N3Lexer);
-    });
-
-    it('should be an N3Lexer constructor', function () {
-      new N3Lexer().should.be.an.instanceof(N3Lexer);
-    });
-
-    it('should provide a shim for setImmediate', function (done) {
-      // Delete global setImmediate
-      var setImmediate = global.setImmediate;
-      global.setImmediate = null;
-      // Clear require cache
-      for (var key in require.cache)
-        delete require.cache[key];
-      // N3Lexer must now fall back to shim
-      var N3LexerWithShim = require('../N3').Lexer;
-      new N3LexerWithShim().tokenize('', done);
-      // Restore global setImmediate
-      global.setImmediate = setImmediate;
+    it('should be an Lexer constructor', () => {
+      new Lexer().should.be.an.instanceof(Lexer);
     });
   });
 
-  describe('An N3Lexer instance', function () {
+  describe('A Lexer instance', () => {
     it('should tokenize the empty string',
       shouldTokenize('',
                      { type: 'eof', line: 1 }));
 
+
+    it('should tokenize byte order mark at beginning as empty string',
+        shouldTokenize('\ufeff', { type: 'eof', line: 1 }));
+
     it('should tokenize a whitespace string',
       shouldTokenize(' \t \n  ',
                      { type: 'eof', line: 2 }));
@@ -45,6 +32,11 @@ describe('N3Lexer', function () {
                      { type: 'IRI', value: 'http://ex.org/?bla#foo', line: 1 },
                      { type: 'eof', line: 1 }));
 
+    it('should tokenize a split IRI',
+      shouldTokenize(streamOf('<', 'http://ex.org/?bla#foo>'),
+        { type: 'IRI', value: 'http://ex.org/?bla#foo', line: 1 },
+        { type: 'eof', line: 1 }));
+
     it('should not tokenize an IRI with disallowed characters',
       shouldNotTokenize('<http://ex.org/bla"foo>',
                         'Unexpected "<http://ex.org/bla"foo>" on line 1.'));
@@ -65,36 +57,24 @@ describe('N3Lexer', function () {
       shouldNotTokenize('<http://ex.org/bla\\uXYZZxyzzfoo>',
                         'Unexpected "<http://ex.org/bla\\uXYZZxyzzfoo>" on line 1.'));
 
-    it('should not tokenize an IRI with a non-numeric 4-digit unicode escapes', function (done) {
-      // Replace global isNaN
-      var isNaN = global.isNaN;
-      global.isNaN = function () { return true; };
-      // Try parsing
-      var stream = new EventEmitter(), lexer = new N3Lexer();
-      lexer.tokenize(stream, function (error, token) {
+    it('should not tokenize an IRI with a non-numeric 4-digit unicode escapes', done => {
+      const stream = new EventEmitter(), lexer = new Lexer();
+      lexer.tokenize(stream, (error, token) => {
         error.should.be.an.instanceof(Error);
-        error.message.should.equal('Unexpected "<\\u1234>" on line 1.');
+        error.message.should.equal('Unexpected "<\\uz234>" on line 1.');
         done(token);
       });
-      stream.emit('data', '<\\u1234>');
-      // Restore global isNaN
-      global.isNaN = isNaN;
+      stream.emit('data', '<\\uz234>');
     });
 
-    it('should not tokenize an IRI with a non-numeric 8-digit unicode escapes', function (done) {
-      // Replace global isNaN
-      var isNaN = global.isNaN;
-      global.isNaN = function () { return true; };
-      // Try parsing
-      var stream = new EventEmitter(), lexer = new N3Lexer();
-      lexer.tokenize(stream, function (error, token) {
+    it('should not tokenize an IRI with a non-numeric 8-digit unicode escapes', done => {
+      const stream = new EventEmitter(), lexer = new Lexer();
+      lexer.tokenize(stream, (error, token) => {
         error.should.be.an.instanceof(Error);
-        error.message.should.equal('Unexpected "<\\U12345678>" on line 1.');
+        error.message.should.equal('Unexpected "<\\Uz2345678>" on line 1.');
         done(token);
       });
-      stream.emit('data', '<\\U12345678>');
-      // Restore global isNaN
-      global.isNaN = isNaN;
+      stream.emit('data', '<\\Uz2345678>');
     });
 
     it('should tokenize an IRI with four-digit unicode escapes',
@@ -260,6 +240,15 @@ describe('N3Lexer', function () {
                      { type: 'IRI', value: '#bla', line: 5 },
                      { type: 'eof', line: 5 }));
 
+    it('should tokenize comments after abbreviations and local names',
+      shouldTokenize('[a#\na:a#\n].',
+                     { type: '[', line: 1 },
+                     { type: 'abbreviation', value: 'a', line: 1 },
+                     { type: 'prefixed', prefix: 'a', value: 'a', line: 2 },
+                     { type: ']', line: 3 },
+                     { type: '.', line: 3 },
+                     { type: 'eof', line: 3 }));
+
     it('should tokenize a quoted string literal',
       shouldTokenize('"string" ',
                      { type: 'literal', value: 'string', line: 1 },
@@ -279,9 +268,9 @@ describe('N3Lexer', function () {
                      { type: 'eof', line: 1 }));
 
     it('should tokenize a triple quoted string literal with quoted newlines inside',
-      shouldTokenize('"""st"r\ni""ng"""',
-                     { type: 'literal', value: 'st"r\ni""ng', line: 1 },
-                     { type: 'eof', line: 2 }));
+      shouldTokenize('"""st"r\r\ni""\n\rng"""',
+                     { type: 'literal', value: 'st"r\r\ni""\n\rng', line: 1 },
+                     { type: 'eof', line: 4 }));
 
     it('should tokenize a string with escape characters',
       shouldTokenize('"\\\\ \\\' \\" \\n \\r \\t \\ua1b2" \n """\\\\ \\\' \\" \\n \\r \\t \\U0000a1b2"""',
@@ -293,6 +282,34 @@ describe('N3Lexer', function () {
       shouldNotTokenize('"\\uXYZX" ',
                         'Unexpected ""\\uXYZX"" on line 1.'));
 
+    it('should not tokenize a double-quoted string ending with an escaped quote',
+      shouldNotTokenize('"abc\\"',
+                        'Unexpected ""abc\\"" on line 1.'));
+
+    it('should not tokenize a single-quoted string ending with an escaped quote',
+      shouldNotTokenize("'abc\\'",
+                        'Unexpected "\'abc\\\'" on line 1.'));
+
+    it('should not tokenize a double-quoted string ending with a backslash and escaped quote',
+      shouldNotTokenize('"abc\\\\\\"',
+                        'Unexpected ""abc\\\\\\"" on line 1.'));
+
+    it('should not tokenize a single-quoted string ending with a backslash and escaped quote',
+      shouldNotTokenize("'abc\\\\\\'",
+                        'Unexpected "\'abc\\\\\\\'" on line 1.'));
+
+    it('should tokenize a double-quoted string ending in backslashes',
+      shouldTokenize('"abc\\\\" "abc\\\\\\\\"',
+                     { type: 'literal', value: 'abc\\', line: 1 },
+                     { type: 'literal', value: 'abc\\\\', line: 1 },
+                     { type: 'eof', line: 1 }));
+
+    it('should tokenize a single-quoted string ending in backslashes',
+      shouldTokenize("'abc\\\\' 'abc\\\\\\\\'",
+                     { type: 'literal', value: 'abc\\', line: 1 },
+                     { type: 'literal', value: 'abc\\\\', line: 1 },
+                     { type: 'eof', line: 1 }));
+
     it('should not tokenize a triple-quoted string with invalid characters',
       shouldNotTokenize("'''\\uXYZX''' ",
                         "Unexpected \"'''\\uXYZX'''\" on line 1."));
@@ -397,13 +414,15 @@ describe('N3Lexer', function () {
                      { type: 'eof', line: 1 }));
 
     it('should tokenize a double literal',
-      shouldTokenize('10e20, +30.40E+50. -60.70e-80. ',
+      shouldTokenize('10e20, +30.40E+50. -60.70e-80. .171e-11.',
                      { type: 'literal', value:      '10e20', prefix: 'http://www.w3.org/2001/XMLSchema#double', line: 1 },
                      { type: ',', line: 1 },
                      { type: 'literal', value: '+30.40E+50', prefix: 'http://www.w3.org/2001/XMLSchema#double', line: 1 },
                      { type: '.', line: 1 },
                      { type: 'literal', value: '-60.70e-80', prefix: 'http://www.w3.org/2001/XMLSchema#double', line: 1 },
                      { type: '.', line: 1 },
+                     { type: 'literal', value: '.171e-11', prefix: 'http://www.w3.org/2001/XMLSchema#double', line: 1 },
+                     { type: '.', line: 1 },
                      { type: 'eof', line: 1 }));
 
     it('should not tokenize an invalid number',
@@ -498,11 +517,33 @@ describe('N3Lexer', function () {
       shouldNotTokenize(streamOf('"abc\r', null), 'Unexpected ""abc" on line 1.'));
 
     it('should tokenize a split single-quoted string',
-      shouldTokenize(streamOf("'abc", "def'"),
+      shouldTokenize(streamOf("'abc", "def' "),
                      { type: 'literal', value: 'abcdef', line: 1 },
                      { type: 'eof', line: 1 }));
 
-    it('should tokenize a split triple-quoted string',
+    it('should tokenize a split single-quoted string followed by a triple-quoted string',
+      // Checks whether _literalClosingPos is properly reset
+      shouldTokenize(streamOf("'abcdef", "' '''a'''"),
+                     { type: 'literal', value: 'abcdef', line: 1 },
+                     { type: 'literal', value: 'a', line: 1 },
+                     { type: 'eof', line: 1 }));
+
+    it('should tokenize split triple-quoted strings',
+      shouldTokenize(streamOf('"', '""abc""" ',
+                              '""', '"def""" ',
+                              '"""', 'ghi""" ',
+                              '"""jkl', '""" ',
+                              '"""mno"', '"" ',
+                              '"""pqr""', '" '),
+                     { type: 'literal', value: 'abc', line: 1 },
+                     { type: 'literal', value: 'def', line: 1 },
+                     { type: 'literal', value: 'ghi', line: 1 },
+                     { type: 'literal', value: 'jkl', line: 1 },
+                     { type: 'literal', value: 'mno', line: 1 },
+                     { type: 'literal', value: 'pqr', line: 1 },
+                     { type: 'eof', line: 1 }));
+
+    it('should tokenize a triple-quoted string split after a newline',
       shouldTokenize(streamOf('"""abc\n', 'def"""'),
                      { type: 'literal', value: 'abc\ndef', line: 1 },
                      { type: 'eof', line: 2 }));
@@ -630,6 +671,41 @@ describe('N3Lexer', function () {
                      { type: '.', line: 1 },
                      { type: 'eof', line: 1 }));
 
+    it('should tokenize the "a" predicate without spacing',
+      shouldTokenize('[a<>].\n[a[]].\n[a()].\n<>a<>.\n<>a[].\n<>a().\n',
+                     { type: '[', line: 1 },
+                     { type: 'abbreviation', value: 'a', line: 1 },
+                     { type: 'IRI', line: 1 },
+                     { type: ']', line: 1 },
+                     { type: '.', line: 1 },
+                     { type: '[', line: 2 },
+                     { type: 'abbreviation', value: 'a', line: 2 },
+                     { type: '[', line: 2 },
+                     { type: ']', line: 2 },
+                     { type: ']', line: 2 },
+                     { type: '.', line: 2 },
+                     { type: '[', line: 3 },
+                     { type: 'abbreviation', value: 'a', line: 3 },
+                     { type: '(', line: 3 },
+                     { type: ')', line: 3 },
+                     { type: ']', line: 3 },
+                     { type: '.', line: 3 },
+                     { type: 'IRI', line: 4 },
+                     { type: 'abbreviation', value: 'a', line: 4 },
+                     { type: 'IRI', line: 4 },
+                     { type: '.', line: 4 },
+                     { type: 'IRI', line: 5 },
+                     { type: 'abbreviation', value: 'a', line: 5 },
+                     { type: '[', line: 5 },
+                     { type: ']', line: 5 },
+                     { type: '.', line: 5 },
+                     { type: 'IRI', line: 6 },
+                     { type: 'abbreviation', value: 'a', line: 6 },
+                     { type: '(', line: 6 },
+                     { type: ')', line: 6 },
+                     { type: '.', line: 6 },
+                     { type: 'eof', line: 7 }));
+
     it('should tokenize an empty default graph',
       shouldTokenize('{}',
                      { type: '{', line: 1 },
@@ -730,6 +806,13 @@ describe('N3Lexer', function () {
                      { type: 'IRI', value: 'b', line: 1 },
                      { type: 'eof', line: 1 }));
 
+    it('should tokenize a split left implication',
+      shouldTokenize(streamOf('<a> <', '= <b> '),
+        { type: 'IRI', value: 'a', line: 1 },
+        { type: 'inverse', value: '>', line: 1 },
+        { type: 'IRI', value: 'b', line: 1 },
+        { type: 'eof', line: 1 }));
+
     it('should tokenize paths',
       shouldTokenize(':joe!fam:mother!loc:office!loc:zip :joe!fam:mother^fam:mother',
                      { type: 'prefixed', prefix: '', value: 'joe', line: 1 },
@@ -750,15 +833,270 @@ describe('N3Lexer', function () {
     it('should not tokenize an invalid document',
       shouldNotTokenize(' \n @!', 'Unexpected "@!" on line 2.'));
 
-    it('does not call setEncoding if not available', function () {
-      new N3Lexer().tokenize({ on: function () {} });
+    it('does not call setEncoding if not available', () => {
+      new Lexer().tokenize({ on: function () {} });
     });
 
-    describe('passing data after the stream has been finished', function () {
-      var tokens = [], error;
-      before(function () {
-        var stream = new EventEmitter(), lexer = new N3Lexer();
-        lexer.tokenize(stream, function (err, token) {
+    it('should tokenize an Quadterm start',
+      shouldTokenize('<<',
+        { type: '<<', line: 1 }, { type: 'eof', line: 1 }));
+
+    it('should tokenize a split Quadterm start',
+      shouldTokenize(streamOf('<', '<'),
+        { type: '<<', line: 1 }, { type: 'eof', line: 1 }));
+
+    it('should tokenize an Quadterm end',
+      shouldTokenize('>>',
+        { type: '>>', line: 1 }, { type: 'eof', line: 1 }));
+
+    it('should tokenize an empty Quadterm',
+      shouldTokenize('<< >>',
+        { type: '<<', line: 1 },
+        { type: '>>', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize an RDF* statement with IRIs',
+      shouldTokenize('<<<http://ex.org/?bla#foo> \n\t<http://ex.org/?bla#bar> \n\t<http://ex.org/?bla#boo>>> .',
+        { type: '<<', line: 1 },
+        { type: 'IRI', value: 'http://ex.org/?bla#foo', line: 1 },
+        { type: 'IRI', value: 'http://ex.org/?bla#bar', line: 2 },
+        { type: 'IRI', value: 'http://ex.org/?bla#boo', line: 3 },
+        { type: '>>', line: 3 },
+        { type: '.', line: 3 },
+        { type: 'eof', line: 3 }));
+
+    it('should not tokenize a wrongly closed RDF* statement with IRIs',
+      shouldNotTokenize('<<<http://ex.org/?bla#foo> \n\t<http://ex.org/?bla#bar> \n\t<http://ex.org/?bla#boo>> .',
+        'Unexpected ">" on line 3.'));
+
+    it('should tokenize a split RDF* statement with IRIs',
+      shouldTokenize(streamOf('<', '<<http://ex.org/?bla#foo> \n\t<http://ex.org/?bla#bar> \n\t<http://ex.org/?bla#boo>>> .'),
+        { type: '<<', line: 1 },
+        { type: 'IRI', value: 'http://ex.org/?bla#foo', line: 1 },
+        { type: 'IRI', value: 'http://ex.org/?bla#bar', line: 2 },
+        { type: 'IRI', value: 'http://ex.org/?bla#boo', line: 3 },
+        { type: '>>', line: 3 },
+        { type: '.', line: 3 },
+        { type: 'eof', line: 3 }));
+
+    it('should tokenize an RDF* statement with string literals',
+      shouldTokenize('<<"string"@en "string"@nl-be "string"@EN>> .',
+        { type: '<<', line: 1 },
+        { type: 'literal', value: 'string', line: 1 },
+        { type: 'langcode', value: 'en', line: 1 },
+        { type: 'literal', value: 'string', line: 1 },
+        { type: 'langcode', value: 'nl-be', line: 1 },
+        { type: 'literal', value: 'string', line: 1 },
+        { type: 'langcode', value: 'EN', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize an RDF* statement with integers',
+      shouldTokenize('<<1 2 3>>.',
+        { type: '<<', line: 1 },
+        { type: 'literal', value: '1', prefix: 'http://www.w3.org/2001/XMLSchema#integer', line: 1 },
+        { type: 'literal', value: '2', prefix: 'http://www.w3.org/2001/XMLSchema#integer', line: 1 },
+        { type: 'literal', value: '3', prefix: 'http://www.w3.org/2001/XMLSchema#integer', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize an RDF* statement with decimals',
+      shouldTokenize('<<1.2 3.4 5.6>>.',
+        { type: '<<', line: 1 },
+        { type: 'literal', value: '1.2', prefix: 'http://www.w3.org/2001/XMLSchema#decimal', line: 1 },
+        { type: 'literal', value: '3.4', prefix: 'http://www.w3.org/2001/XMLSchema#decimal', line: 1 },
+        { type: 'literal', value: '5.6', prefix: 'http://www.w3.org/2001/XMLSchema#decimal', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize an RDF* statement with booleans',
+      shouldTokenize('<<true false true>>.',
+        { type: '<<', line: 1 },
+        { type: 'literal', value: 'true',  prefix: 'http://www.w3.org/2001/XMLSchema#boolean', line: 1 },
+        { type: 'literal', value: 'false', prefix: 'http://www.w3.org/2001/XMLSchema#boolean', line: 1 },
+        { type: 'literal', value: 'true',  prefix: 'http://www.w3.org/2001/XMLSchema#boolean', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize a prefixed iri followed by the end of a QuadTerm',
+      shouldTokenize('c:c>> .',
+        { type: 'prefixed', prefix: 'c', value: 'c', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize an RDF* statement with prefixed names',
+      shouldTokenize('<<a:a b:b c:c>> .',
+        { type: '<<', line: 1 },
+        { type: 'prefixed', prefix: 'a', value: 'a', line: 1 },
+        { type: 'prefixed', prefix: 'b', value: 'b', line: 1 },
+        { type: 'prefixed', prefix: 'c', value: 'c', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize an RDF* statement with blank nodes',
+      shouldTokenize('<<_:a _:b _:c>> .',
+        { type: '<<', line: 1 },
+        { type: 'blank', prefix: '_', value: 'a', line: 1 },
+        { type: 'blank', prefix: '_', value: 'b', line: 1 },
+        { type: 'blank', prefix: '_', value: 'c', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize an RDF* statement with variables',
+      shouldTokenize('<<?a ?b ?c>> .',
+        { type: '<<', line: 1 },
+        { type: 'var', value: '?a', line: 1 },
+        { type: 'var', value: '?b', line: 1 },
+        { type: 'var', value: '?c', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize an RDF* statement with mixed types',
+      shouldTokenize('<<<http://ex.org/?bla#foo> "string"@nl-be c:c>> .',
+        { type: '<<', line: 1 },
+        { type: 'IRI', value: 'http://ex.org/?bla#foo', line: 1 },
+        { type: 'literal', value: 'string', line: 1 },
+        { type: 'langcode', value: 'nl-be', line: 1 },
+        { type: 'prefixed', prefix: 'c', value: 'c', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize an RDF* statement with mixed types',
+      shouldTokenize('<<_:a a:a "string"@EN>> .',
+        { type: '<<', line: 1 },
+        { type: 'blank', prefix: '_', value: 'a', line: 1 },
+        { type: 'prefixed', prefix: 'a', value: 'a', line: 1 },
+        { type: 'literal', value: 'string', line: 1 },
+        { type: 'langcode', value: 'EN', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize an RDF* statement with mixed types',
+      shouldTokenize('<<"literal"@AU <http://ex.org/?bla#foo> _:a>> .',
+        { type: '<<', line: 1 },
+        { type: 'literal', value: 'literal', line: 1 },
+        { type: 'langcode', value: 'AU', line: 1 },
+        { type: 'IRI', value: 'http://ex.org/?bla#foo', line: 1 },
+        { type: 'blank', prefix: '_', value: 'a', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize RDF* statements with shared subjects',
+      shouldTokenize('<<<a> <b> <c>;\n<d> <e>>>.',
+        { type: '<<', line: 1 },
+        { type: 'IRI', value: 'a', line: 1 },
+        { type: 'IRI', value: 'b', line: 1 },
+        { type: 'IRI', value: 'c', line: 1 },
+        { type: ';', line: 1 },
+        { type: 'IRI', value: 'd', line: 2 },
+        { type: 'IRI', value: 'e', line: 2 },
+        { type: '>>', line: 2 },
+        { type: '.', line: 2 },
+        { type: 'eof', line: 2 }));
+
+    it('should tokenize RDF* statements with shared subjects and predicates',
+      shouldTokenize('<<<a> <b> <c>,\n<d>>>.',
+        { type: '<<', line: 1 },
+        { type: 'IRI', value: 'a', line: 1 },
+        { type: 'IRI', value: 'b', line: 1 },
+        { type: 'IRI', value: 'c', line: 1 },
+        { type: ',', line: 1 },
+        { type: 'IRI', value: 'd', line: 2 },
+        { type: '>>', line: 2 },
+        { type: '.', line: 2 },
+        { type: 'eof', line: 2 }));
+
+    it('should tokenize an RDF* statement with shared subjects and predicates and prefixed names',
+      shouldTokenize('<<a:a b:b c:c;d:d e:e,f:f>> .',
+        { type: '<<', line: 1 },
+        { type: 'prefixed', prefix: 'a', value: 'a', line: 1 },
+        { type: 'prefixed', prefix: 'b', value: 'b', line: 1 },
+        { type: 'prefixed', prefix: 'c', value: 'c', line: 1 },
+        { type: ';', line: 1 },
+        { type: 'prefixed', prefix: 'd', value: 'd', line: 1 },
+        { type: 'prefixed', prefix: 'e', value: 'e', line: 1 },
+        { type: ',', line: 1 },
+        { type: 'prefixed', prefix: 'f', value: 'f', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize a QuadTerm followed by other tokens',
+      shouldTokenize('<<_:a <b> "lit"@EN>> _:a b:b.',
+        { type: '<<', line: 1 },
+        { type: 'blank', prefix: '_', value: 'a', line: 1 },
+        { type: 'IRI', value: 'b', line: 1 },
+        { type: 'literal', value: 'lit', line: 1 },
+        { type: 'langcode', value: 'EN', line: 1 },
+        { type: '>>', line: 1 },
+        { type: 'blank', prefix: '_', value: 'a', line: 1 },
+        { type: 'prefixed', prefix: 'b', value: 'b', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize a QuadTerm preceded by other tokens',
+      shouldTokenize('"lit"@DE _:b <<_:a b:b "lit"@EN>>.',
+        { type: 'literal', value: 'lit', line: 1 },
+        { type: 'langcode', value: 'DE', line: 1 },
+        { type: 'blank', prefix: '_', value: 'b', line: 1 },
+        { type: '<<', line: 1 },
+        { type: 'blank', prefix: '_', value: 'a', line: 1 },
+        { type: 'prefixed', prefix: 'b', value: 'b', line: 1 },
+        { type: 'literal', value: 'lit', line: 1 },
+        { type: 'langcode', value: 'EN', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize a nested QuadTerm as subject in a statement',
+      shouldTokenize('<<<<_:b <b> "lit"@DE>> <a> "lit"@EN>>.',
+        { type: '<<', line: 1 },
+        { type: '<<', line: 1 },
+        { type: 'blank', prefix: '_', value: 'b', line: 1 },
+        { type: 'IRI', value: 'b', line: 1 },
+        { type: 'literal', value: 'lit', line: 1 },
+        { type: 'langcode', value: 'DE', line: 1 },
+        { type: '>>', line: 1 },
+        { type: 'IRI', value: 'a', line: 1 },
+        { type: 'literal', value: 'lit', line: 1 },
+        { type: 'langcode', value: 'EN', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+    it('should tokenize a nested QuadTerm as object in a statement',
+      shouldTokenize('<<a:a _:a <<_:b <b> "lit"@DE>>>>.',
+        { type: '<<', line: 1 },
+        { type: 'prefixed', prefix: 'a', value: 'a', line: 1 },
+        { type: 'blank', prefix: '_', value: 'a', line: 1 },
+        { type: '<<', line: 1 },
+        { type: 'blank', prefix: '_', value: 'b', line: 1 },
+        { type: 'IRI', value: 'b', line: 1 },
+        { type: 'literal', value: 'lit', line: 1 },
+        { type: 'langcode', value: 'DE', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '>>', line: 1 },
+        { type: '.', line: 1 },
+        { type: 'eof', line: 1 }));
+
+
+    describe('passing data after the stream has been finished', () => {
+      const tokens = [];
+      let error;
+      before(() => {
+        const stream = new EventEmitter(), lexer = new Lexer();
+        lexer.tokenize(stream, (err, token) => {
           if (err)
             error = err;
           else
@@ -770,20 +1108,21 @@ describe('N3Lexer', function () {
         stream.emit('end');
       });
 
-      it('parses only the first chunk (plus EOF)', function () {
+      it('parses only the first chunk (plus EOF)', () => {
         tokens.should.have.length(2);
       });
 
-      it('does not emit an error', function () {
+      it('does not emit an error', () => {
         expect(error).to.not.exist;
       });
     });
 
-    describe('passing data after a syntax error', function () {
-      var tokens = [], error;
-      before(function () {
-        var stream = new EventEmitter(), lexer = new N3Lexer();
-        lexer.tokenize(stream, function (err, token) {
+    describe('passing data after a syntax error', () => {
+      const tokens = [];
+      let error;
+      before(() => {
+        const stream = new EventEmitter(), lexer = new Lexer();
+        lexer.tokenize(stream, (err, token) => {
           if (err)
             error = err;
           else
@@ -796,21 +1135,22 @@ describe('N3Lexer', function () {
         stream.emit('end');
       });
 
-      it('parses only the first chunk', function () {
+      it('parses only the first chunk', () => {
         tokens.should.have.length(1);
       });
 
-      it('emits the error', function () {
+      it('emits the error', () => {
         expect(error).to.exist;
         expect(error).to.have.property('message', 'Unexpected "error" on line 1.');
       });
     });
 
-    describe('when the stream errors', function () {
-      var tokens = [], error;
-      before(function () {
-        var stream = new EventEmitter(), lexer = new N3Lexer();
-        lexer.tokenize(stream, function (err, token) {
+    describe('when the stream errors', () => {
+      const tokens = [];
+      let error;
+      before(() => {
+        const stream = new EventEmitter(), lexer = new Lexer();
+        lexer.tokenize(stream, (err, token) => {
           if (err)
             error = err;
           else
@@ -819,17 +1159,17 @@ describe('N3Lexer', function () {
         stream.emit('error', new Error('my error'));
       });
 
-      it('passes the error', function () {
+      it('passes the error', () => {
         expect(error).to.exist;
         expect(error).to.have.property('message', 'my error');
       });
     });
 
-    describe('called with a string and without callback', function () {
-      var lexer = new N3Lexer(),
+    describe('called with a string and without callback', () => {
+      const lexer = new Lexer(),
           tokens = lexer.tokenize('<a> <b> <c>.');
 
-      it('returns all tokens synchronously', function () {
+      it('returns all tokens synchronously', () => {
         tokens.should.deep.equal([
           { line: 1, type: 'IRI', value: 'a', prefix: '' },
           { line: 1, type: 'IRI', value: 'b', prefix: '' },
@@ -840,10 +1180,10 @@ describe('N3Lexer', function () {
       });
     });
 
-    describe('called with an erroneous string and without callback', function () {
-      var lexer = new N3Lexer();
+    describe('called with an erroneous string and without callback', () => {
+      const lexer = new Lexer();
 
-      it('throws an error', function () {
+      it('throws an error', () => {
         (function () { lexer.tokenize('<a> bar'); })
         .should.throw('Unexpected "bar" on line 1.');
       });
@@ -851,8 +1191,8 @@ describe('N3Lexer', function () {
   });
 });
 
-describe('An N3Lexer instance with the n3 option set to false', function () {
-  function createLexer() { return new N3Lexer({ n3: false }); }
+describe('A Lexer instance with the n3 option set to false', () => {
+  function createLexer() { return new Lexer({ n3: false }); }
 
   it('should not tokenize a variable',
     shouldNotTokenize(createLexer(), '?a', 'Unexpected "?a" on line 1.'));
@@ -873,8 +1213,8 @@ describe('An N3Lexer instance with the n3 option set to false', function () {
     shouldNotTokenize(createLexer(), ':joe^fam:father', 'Unexpected "^fam:father" on line 1.'));
 });
 
-describe('An N3Lexer instance with the comment option set to true', function () {
-  function createLexer() { return new N3Lexer({ comments: true }); }
+describe('A Lexer instance with the comment option set to true', () => {
+  function createLexer() { return new Lexer({ comments: true }); }
 
   it('should tokenize a single comment',
     shouldTokenize(createLexer(), '#mycomment',
@@ -899,23 +1239,23 @@ describe('An N3Lexer instance with the comment option set to true', function ()
 });
 
 function shouldTokenize(lexer, input) {
-  var expected = Array.prototype.slice.call(arguments, 1);
+  const expected = Array.prototype.slice.call(arguments, 1);
   // Shift parameters as necessary
-  if (lexer instanceof N3Lexer)
+  if (lexer instanceof Lexer)
     expected.shift();
   else
-    input = lexer, lexer = new N3Lexer();
+    input = lexer, lexer = new Lexer();
 
   return function (done) {
-    var result = [];
+    const result = [];
     lexer.tokenize(input, tokenCallback);
 
     function tokenCallback(error, token) {
       expect(error).not.to.exist;
       expect(token).to.exist;
-      var expectedItem = expected[result.length];
+      const expectedItem = expected[result.length];
       if (expectedItem)
-        for (var attribute in token)
+        for (const attribute in token)
           if (token[attribute] === '' && expectedItem[attribute] !== '')
             delete token[attribute];
       result.push(token);
@@ -929,8 +1269,8 @@ function shouldTokenize(lexer, input) {
 
 function shouldNotTokenize(lexer, input, expectedError) {
   // Shift parameters if necessary
-  if (!(lexer instanceof N3Lexer))
-    expectedError = input, input = lexer, lexer = new N3Lexer();
+  if (!(lexer instanceof Lexer))
+    expectedError = input, input = lexer, lexer = new Lexer();
 
   return function (done) {
     lexer.tokenize(input, tokenCallback);
@@ -942,30 +1282,27 @@ function shouldNotTokenize(lexer, input, expectedError) {
         done();
       }
       else if (token.type === 'eof')
-        throw new Error('Expected error ' + expectedError);
+        throw new Error(`Expected error ${expectedError}`);
     }
   };
 }
 
-var immediately = typeof setImmediate === 'function' ? setImmediate :
-                  function setImmediate(func) { setTimeout(func, 0); };
-
 function streamOf() {
-  var elements = Array.prototype.slice.call(arguments),
+  const elements = Array.prototype.slice.call(arguments),
       stream = new EventEmitter();
 
   stream.setEncoding = function (encoding) {
     if (encoding === 'utf8')
-      immediately(next, 0);
+      queueMicrotask(next);
   };
 
   function next() {
     if (elements.length) {
-      var element = elements.shift();
+      const element = elements.shift();
       // use "null" to stall the stream
       if (element !== null) {
         stream.emit('data', element);
-        immediately(next, 0);
+        queueMicrotask(next);
       }
     }
     else {
