@@ -1,40 +1,40 @@
-var N3Writer = require('../N3').Writer;
+import {
+  Writer,
+  NamedNode,
+  BlankNode,
+  Literal,
+  Quad,
+  termFromId,
+} from '../src/';
+import namespaces from '../src/IRIs';
 
-var DataFactory = require('../N3').DataFactory;
-var NamedNode = DataFactory.internal.NamedNode,
-    Literal = DataFactory.internal.Literal,
-    Quad = DataFactory.internal.Quad,
-    fromId = DataFactory.internal.fromId;
+const { xsd } = namespaces;
 
-describe('N3Writer', function () {
-  describe('The N3Writer module', function () {
-    it('should be a function', function () {
-      N3Writer.should.be.a('function');
+describe('Writer', () => {
+  describe('The Writer export', () => {
+    it('should be a function', () => {
+      Writer.should.be.a('function');
     });
 
-    it('should make N3Writer objects', function () {
-      N3Writer().should.be.an.instanceof(N3Writer);
-    });
-
-    it('should be an N3Writer constructor', function () {
-      new N3Writer().should.be.an.instanceof(N3Writer);
+    it('should be a Writer constructor', () => {
+      new Writer().should.be.an.instanceof(Writer);
     });
   });
 
-  describe('An N3Writer instance', function () {
-    it('should serialize a single triple', function () {
-      var writer = N3Writer();
+  describe('A Writer instance', () => {
+    it('should serialize a single triple', () => {
+      const writer = new Writer();
       writer.quadToString(new NamedNode('a'), new NamedNode('b'), new NamedNode('c')).should.equal('<a> <b> <c> .\n');
     });
 
-    it('should serialize a single quad', function () {
-      var writer = N3Writer();
+    it('should serialize a single quad', () => {
+      const writer = new Writer();
       writer.quadToString(new NamedNode('a'), new NamedNode('b'), new NamedNode('c'), new NamedNode('g')).should.equal('<a> <b> <c> <g> .\n');
     });
 
-    it('should serialize an array of triples', function () {
-      var writer = N3Writer();
-      var triples = [new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c')),
+    it('should serialize an array of triples', () => {
+      const writer = new Writer();
+      const triples = [new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c')),
         new Quad(new NamedNode('d'), new NamedNode('e'), new NamedNode('f'))];
       writer.quadsToString(triples).should.equal('<a> <b> <c> .\n<d> <e> <f> .\n');
     });
@@ -122,6 +122,66 @@ describe('N3Writer', function () {
       shouldSerialize(['a', 'b', '"c\u0000\u0001"'],
                       '<a> <b> "c\\u0000\\u0001".\n'));
 
+    it('should serialize a true boolean literal',
+      shouldSerialize(['a', 'b', `"true"^^${xsd.boolean}`],
+                      '<a> <b> true.\n'));
+
+    it('should serialize a false boolean literal',
+      shouldSerialize(['a', 'b', `"false"^^${xsd.boolean}`],
+                      '<a> <b> false.\n'));
+
+    it('should serialize an invalid boolean literal',
+      shouldSerialize(['a', 'b', `"invalid"^^${xsd.boolean}`],
+                      `<a> <b> "invalid"^^<${xsd.boolean}>.\n`));
+
+    it('should serialize an integer literal',
+      shouldSerialize(['a', 'b', `"123"^^${xsd.integer}`],
+                      '<a> <b> 123.\n'));
+
+    it('should serialize a positive integer literal',
+      shouldSerialize(['a', 'b', `"+123"^^${xsd.integer}`],
+                      '<a> <b> +123.\n'));
+
+    it('should serialize a negative integer literal',
+      shouldSerialize(['a', 'b', `"-123"^^${xsd.integer}`],
+                      '<a> <b> -123.\n'));
+
+    it('should serialize an invalid integer literal',
+      shouldSerialize(['a', 'b', `"invalid"^^${xsd.integer}`],
+                      `<a> <b> "invalid"^^<${xsd.integer}>.\n`));
+
+    it('should serialize a decimal literal',
+      shouldSerialize(['a', 'b', `"123.456"^^${xsd.decimal}`],
+                      '<a> <b> 123.456.\n'));
+
+    it('should serialize a positive decimal literal',
+      shouldSerialize(['a', 'b', `"+123.456"^^${xsd.decimal}`],
+                      '<a> <b> +123.456.\n'));
+
+    it('should serialize a negative decimal literal',
+      shouldSerialize(['a', 'b', `"-123.456"^^${xsd.decimal}`],
+                      '<a> <b> -123.456.\n'));
+
+    it('should serialize an invalid decimal literal',
+      shouldSerialize(['a', 'b', `"invalid"^^${xsd.decimal}`],
+                      `<a> <b> "invalid"^^<${xsd.decimal}>.\n`));
+
+    it('should serialize a double literal',
+      shouldSerialize(['a', 'b', `"123.456E10"^^${xsd.double}`],
+                      '<a> <b> 123.456E10.\n'));
+
+    it('should serialize a positive double literal',
+      shouldSerialize(['a', 'b', `"+123.456E10"^^${xsd.double}`],
+                      '<a> <b> +123.456E10.\n'));
+
+    it('should serialize a negative double literal',
+      shouldSerialize(['a', 'b', `"-123.456E10"^^${xsd.double}`],
+                      '<a> <b> -123.456E10.\n'));
+
+    it('should serialize an invalid double literal',
+      shouldSerialize(['a', 'b', `"invalid"^^${xsd.double}`],
+                      `<a> <b> "invalid"^^<${xsd.double}>.\n`));
+
     it('should serialize blank nodes',
       shouldSerialize(['_:a', 'b', { termType: 'BlankNode', value: 'c' }],
                       '_:a <b> _:c.\n'));
@@ -131,10 +191,11 @@ describe('N3Writer', function () {
                       ['a', 'b', 'c'],
                       '<a> <b> <c>.\n'));
 
-    it('should serialize valid prefixes',
-      shouldSerialize({ prefixes: { a: 'http://a.org/', b: new NamedNode('http://a.org/b#'), c: 'http://a.org/b' } },
+    it('should serialize prefixes',
+      shouldSerialize({ prefixes: { a: 'http://a.org/', b: new NamedNode('http://a.org/b#'), c: 'http://a.org/c' } },
                       '@prefix a: <http://a.org/>.\n' +
-                      '@prefix b: <http://a.org/b#>.\n\n'));
+                      '@prefix b: <http://a.org/b#>.\n' +
+                      '@prefix c: <http://a.org/c>.\n\n'));
 
     it('should use prefixes when possible',
       shouldSerialize({ prefixes: { a: 'http://a.org/', b: 'http://a.org/b#', c: 'http://a.org/b' } },
@@ -142,7 +203,8 @@ describe('N3Writer', function () {
                       ['http://a.org/bc/de', 'http://a.org/b#e#f', 'http://a.org/b#x/t'],
                       ['http://a.org/3a', 'http://a.org/b#3a', 'http://a.org/b#a3'],
                       '@prefix a: <http://a.org/>.\n' +
-                      '@prefix b: <http://a.org/b#>.\n\n' +
+                      '@prefix b: <http://a.org/b#>.\n' +
+                      '@prefix c: <http://a.org/b>.\n\n' +
                       'a:bc b:ef a:bhi.\n' +
                       '<http://a.org/bc/de> <http://a.org/b#e#f> <http://a.org/b#x/t>.\n' +
                       '<http://a.org/3a> <http://a.org/b#3a> b:a3.\n'));
@@ -208,98 +270,134 @@ describe('N3Writer', function () {
       shouldSerialize(['_:\ud835\udc00', '_:\ud835\udc00', '_:\ud835\udc00', '_:\ud835\udc00'],
                       '_:\ud835\udc00 {\n_:\ud835\udc00 _:\ud835\udc00 _:\ud835\udc00\n}\n'));
 
-    it('calls the done callback when ending the outputstream errors', function (done) {
-      var writer = new N3Writer({
+    it('calls the done callback when ending the outputstream errors', done => {
+      const writer = new Writer({
         write: function () {},
         end: function () { throw new Error('error'); },
       });
       writer.end(done);
     });
 
-    it('sends output through end when no stream argument is given', function (done) {
-      var writer = new N3Writer(), notCalled = true;
-      writer.addQuad(new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c')), function () { notCalled = false; });
-      writer.end(function (error, output) {
+    it('sends output through end when no stream argument is given', done => {
+      const writer = new Writer();
+      let notCalled = true;
+      writer.addQuad(new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c')), () => { notCalled = false; });
+      writer.end((error, output) => {
         output.should.equal('<a> <b> <c>.\n');
         done(notCalled || error);
       });
     });
 
-    it('respects the prefixes argument when no stream argument is given', function (done) {
-      var writer = new N3Writer({ prefixes: { a: 'b#' } });
+    it('respects the prefixes argument when no stream argument is given', done => {
+      const writer = new Writer({ prefixes: { a: 'b#' } });
       writer.addQuad(new Quad(new NamedNode('b#a'), new NamedNode('b#b'), new NamedNode('b#c')));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('@prefix a: <b#>.\n\na:a a:b a:c.\n');
         done(error);
       });
     });
 
-    it('does not repeat identical prefixes', function (done) {
-      var writer = new N3Writer();
-      writer.addPrefix('a', 'b#');
-      writer.addPrefix('a', 'b#');
+    it('ignores an empty prefix list', done => {
+      const writer = new Writer();
+      writer.addPrefixes({});
       writer.addQuad(new Quad(new NamedNode('b#a'), new NamedNode('b#b'), new NamedNode('b#c')));
-      writer.addPrefix('a', 'b#');
-      writer.addPrefix('a', 'b#');
-      writer.addPrefix('b', 'b#');
-      writer.addPrefix('a', 'c#');
-      writer.end(function (error, output) {
-        output.should.equal('@prefix a: <b#>.\n\na:a a:b a:c.\n' +
-                            '@prefix b: <b#>.\n\n@prefix a: <c#>.\n\n');
+      writer.end((error, output) => {
+        output.should.equal('<b#a> <b#b> <b#c>.\n');
         done(error);
       });
     });
 
-    it('serializes triples of a graph with a prefix declaration in between', function (done) {
-      var writer = new N3Writer();
+    it('serializes triples of a graph with a prefix declaration in between', done => {
+      const writer = new Writer();
+      writer.addQuad(new Quad(new NamedNode('b#a'), new NamedNode('b#b'), new NamedNode('b#c')));
       writer.addPrefix('a', 'b#');
       writer.addQuad(new Quad(new NamedNode('b#a'), new NamedNode('b#b'), new NamedNode('b#c'), new NamedNode('b#g')));
       writer.addPrefix('d', 'e#');
-      writer.addQuad(new Quad(new NamedNode('b#a'), new NamedNode('b#b'), new NamedNode('b#d'), new NamedNode('b#g')));
-      writer.end(function (error, output) {
-        output.should.equal('@prefix a: <b#>.\n\na:g {\na:a a:b a:c\n}\n' +
+      writer.addQuad({ subject: new NamedNode('b#a'), predicate: new NamedNode('b#b'), object: new NamedNode('b#d'), graph: new NamedNode('b#g') });
+      writer.end((error, output) => {
+        output.should.equal('<b#a> <b#b> <b#c>.\n' +
+                            '@prefix a: <b#>.\n\na:g {\na:a a:b a:c\n}\n' +
                             '@prefix d: <e#>.\n\na:g {\na:a a:b a:d\n}\n');
         done(error);
       });
     });
 
-    it('should accept triples with separated components', function (done) {
-      var writer = N3Writer();
+    it('should not write prefixes in N-Triples mode', done => {
+      const writer = new Writer({ format: 'N-Triples', prefixes: { a: 'b#' } });
+      let called = false;
+      function callback() { called = true; }
+      writer.addPrefix('c', 'd#');
+      writer.addQuad(new NamedNode('a'), new NamedNode('b'), new Literal('"c"'));
+      writer.addQuad(new NamedNode('a'), new NamedNode('b'), new Literal(`"1"^^${xsd.integer}`));
+      writer.addPrefix('e', 'f#', callback);
+      writer.end((error, output) => {
+        called.should.be.true;
+        output.should.equal(`<a> <b> "c" .\n<a> <b> "1"^^<${xsd.integer}> .\n`);
+        done(error);
+      });
+    });
+
+    it('uses a base IRI when given', done => {
+      const writer = new Writer({ baseIRI: 'http://example.org/foo/' });
+      writer.addQuad(new Quad(
+        new NamedNode('http://example.org/foo/'),
+        new NamedNode('http://example.org/foo/#b'),
+        new NamedNode('http://example.org/foo/cdeFgh/ijk')));
+      writer.end((error, output) => {
+        output.should.equal('<> <#b> <cdeFgh/ijk>.\n');
+        done(error);
+      });
+    });
+
+    it('does not use partially match base IRIs', done => {
+      const writer = new Writer({ baseIRI: 'https://pod.example/profile/card' });
+      writer.addQuad(new Quad(
+        new NamedNode('https://pod.example/profile/card#me'),
+        new NamedNode('http://www.w3.org/2002/07/owl#sameAs'),
+        new NamedNode('https://pod.example/profile/card-1234.ttl')));
+      writer.end((error, output) => {
+        output.should.equal('<#me> <http://www.w3.org/2002/07/owl#sameAs> <https://pod.example/profile/card-1234.ttl>.\n');
+        done(error);
+      });
+    });
+
+    it('should accept triples with separated components', done => {
+      const writer = new Writer();
       writer.addQuad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c'));
       writer.addQuad(new NamedNode('a'), new NamedNode('b'), new NamedNode('d'));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('<a> <b> <c>, <d>.\n');
         done(error);
       });
     });
 
-    it('should accept quads with separated components', function (done) {
-      var writer = N3Writer();
+    it('should accept quads with separated components', done => {
+      const writer = new Writer();
       writer.addQuad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c'), new NamedNode('g'));
       writer.addQuad(new NamedNode('a'), new NamedNode('b'), new NamedNode('d'), new NamedNode('g'));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('<g> {\n<a> <b> <c>, <d>\n}\n');
         done(error);
       });
     });
 
-    it('should serialize triples with an empty blank node as object', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with an empty blank node as object', done => {
+      const writer = new Writer();
       writer.addQuad(new NamedNode('a1'), new NamedNode('b'), writer.blank());
       writer.addQuad(new NamedNode('a2'), new NamedNode('b'), writer.blank([]));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('<a1> <b> [].\n' +
                             '<a2> <b> [].\n');
         done(error);
       });
     });
 
-    it('should serialize triples with a one-triple blank node as object', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with a one-triple blank node as object', done => {
+      const writer = new Writer();
       writer.addQuad(new NamedNode('a1'), new NamedNode('b'), writer.blank(new NamedNode('d'), new NamedNode('e')));
       writer.addQuad(new NamedNode('a2'), new NamedNode('b'), writer.blank({ predicate: new NamedNode('d'), object: new NamedNode('e') }));
       writer.addQuad(new NamedNode('a3'), new NamedNode('b'), writer.blank([{ predicate: new NamedNode('d'), object: new NamedNode('e') }]));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('<a1> <b> [ <d> <e> ].\n' +
                             '<a2> <b> [ <d> <e> ].\n' +
                             '<a3> <b> [ <d> <e> ].\n');
@@ -307,13 +405,13 @@ describe('N3Writer', function () {
       });
     });
 
-    it('should serialize triples with a two-triple blank node as object', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with a two-triple blank node as object', done => {
+      const writer = new Writer();
       writer.addQuad(new NamedNode('a'), new NamedNode('b'), writer.blank([
           { predicate: new NamedNode('d'), object: new NamedNode('e') },
           { predicate: new NamedNode('f'), object: new Literal('"g"') },
       ]));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('<a> <b> [\n' +
                             '  <d> <e>;\n' +
                             '  <f> "g"\n' +
@@ -322,14 +420,14 @@ describe('N3Writer', function () {
       });
     });
 
-    it('should serialize triples with a three-triple blank node as object', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with a three-triple blank node as object', done => {
+      const writer = new Writer();
       writer.addQuad(new NamedNode('a'), new NamedNode('b'), writer.blank([
         { predicate: new NamedNode('d'), object: new NamedNode('e') },
         { predicate: new NamedNode('f'), object: new Literal('"g"') },
         { predicate: new NamedNode('h'), object: new NamedNode('i') },
       ]));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('<a> <b> [\n' +
                             '  <d> <e>;\n' +
                             '  <f> "g";\n' +
@@ -339,15 +437,15 @@ describe('N3Writer', function () {
       });
     });
 
-    it('should serialize triples with predicate-sharing blank node triples as object', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with predicate-sharing blank node triples as object', done => {
+      const writer = new Writer();
       writer.addQuad(new NamedNode('a'), new NamedNode('b'), writer.blank([
         { predicate: new NamedNode('d'), object: new NamedNode('e') },
         { predicate: new NamedNode('d'), object: new NamedNode('f') },
         { predicate: new NamedNode('g'), object: new NamedNode('h') },
         { predicate: new NamedNode('g'), object: new NamedNode('i') },
       ]));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('<a> <b> [\n' +
           '  <d> <e>, <f>;\n' +
           '  <g> <h>, <i>\n' +
@@ -356,8 +454,8 @@ describe('N3Writer', function () {
       });
     });
 
-    it('should serialize triples with nested blank nodes as object', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with nested blank nodes as object', done => {
+      const writer = new Writer();
       writer.addQuad(new NamedNode('a1'), new NamedNode('b'), writer.blank([
         { predicate: new NamedNode('d'), object: writer.blank() },
       ]));
@@ -371,7 +469,7 @@ describe('N3Writer', function () {
           { predicate: new NamedNode('j'), object: writer.blank(new NamedNode('k'), new Literal('"l"')) },
         ]) },
       ]));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('<a1> <b> [\n' +
           '  <d> []\n' +
           '].\n' +
@@ -389,23 +487,23 @@ describe('N3Writer', function () {
       });
     });
 
-    it('should serialize triples with an empty blank node as subject', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with an empty blank node as subject', done => {
+      const writer = new Writer();
       writer.addQuad(writer.blank(), new NamedNode('b'), new NamedNode('c'));
       writer.addQuad(writer.blank([]), new NamedNode('b'), new NamedNode('c'));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('[] <b> <c>.\n' +
                             '[] <b> <c>.\n');
         done(error);
       });
     });
 
-    it('should serialize triples with a one-triple blank node as subject', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with a one-triple blank node as subject', done => {
+      const writer = new Writer();
       writer.addQuad(writer.blank(new NamedNode('a'), new NamedNode('b')), new NamedNode('c'), new NamedNode('d'));
       writer.addQuad(writer.blank({ predicate: new NamedNode('a'), object: new NamedNode('b') }), new NamedNode('c'), new NamedNode('d'));
       writer.addQuad(writer.blank([{ predicate: new NamedNode('a'), object: new NamedNode('b') }]), new NamedNode('c'), new NamedNode('d'));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('[ <a> <b> ] <c> <d>.\n' +
                             '[ <a> <b> ] <c> <d>.\n' +
                             '[ <a> <b> ] <c> <d>.\n');
@@ -413,173 +511,249 @@ describe('N3Writer', function () {
       });
     });
 
-    it('should serialize triples with an empty blank node as graph', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with an empty blank node as graph', done => {
+      const writer = new Writer();
       writer.addQuad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c'), writer.blank());
       writer.addQuad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c'), writer.blank([]));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('[] {\n<a> <b> <c>\n}\n' +
                             '[] {\n<a> <b> <c>\n}\n');
         done(error);
       });
     });
 
-    it('should serialize triples with an empty list as object', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with an empty list as object', done => {
+      const writer = new Writer();
       writer.addQuad(new NamedNode('a1'), new NamedNode('b'), writer.list());
       writer.addQuad(new NamedNode('a2'), new NamedNode('b'), writer.list([]));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('<a1> <b> ().\n' +
                             '<a2> <b> ().\n');
         done(error);
       });
     });
 
-    it('should serialize triples with a one-element list as object', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with a one-element list as object', done => {
+      const writer = new Writer();
       writer.addQuad(new NamedNode('a1'), new NamedNode('b'), writer.list([new NamedNode('c')]));
       writer.addQuad(new NamedNode('a2'), new NamedNode('b'), writer.list([new Literal('"c"')]));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('<a1> <b> (<c>).\n' +
                             '<a2> <b> ("c").\n');
         done(error);
       });
     });
 
-    it('should serialize triples with a three-element list as object', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with a three-element list as object', done => {
+      const writer = new Writer();
       writer.addQuad(new NamedNode('a1'), new NamedNode('b'), writer.list([new NamedNode('c'), new NamedNode('d'), new NamedNode('e')]));
       writer.addQuad(new NamedNode('a2'), new NamedNode('b'), writer.list([new Literal('"c"'), new Literal('"d"'), new Literal('"e"')]));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('<a1> <b> (<c> <d> <e>).\n' +
                             '<a2> <b> ("c" "d" "e").\n');
         done(error);
       });
     });
 
-    it('should serialize triples with an empty list as subject', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with an empty list as subject', done => {
+      const writer = new Writer();
       writer.addQuad(writer.list(),   new NamedNode('b1'), new NamedNode('c'));
       writer.addQuad(writer.list([]), new NamedNode('b2'), new NamedNode('c'));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('() <b1> <c>.\n' +
                             '() <b2> <c>.\n');
         done(error);
       });
     });
 
-    it('should serialize triples with a one-element list as subject', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with a one-element list as subject', done => {
+      const writer = new Writer();
       writer.addQuad(writer.list([new NamedNode('a')]), new NamedNode('b1'), new NamedNode('c'));
       writer.addQuad(writer.list([new NamedNode('a')]), new NamedNode('b2'), new NamedNode('c'));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('(<a>) <b1> <c>.\n' +
                             '(<a>) <b2> <c>.\n');
         done(error);
       });
     });
 
-    it('should serialize triples with a three-element list as subject', function (done) {
-      var writer = N3Writer();
+    it('should serialize triples with a three-element list as subject', done => {
+      const writer = new Writer();
       writer.addQuad(writer.list([new NamedNode('a1'), new Literal('"b"'), new Literal('"c"')]), new NamedNode('d'), new NamedNode('e'));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('(<a1> "b" "c") <d> <e>.\n');
         done(error);
       });
     });
 
-    it('should accept triples in bulk', function (done) {
-      var writer = N3Writer();
+    it('should serialize subject and object triples passed by options.listHeads', done => {
+      const lists = {
+        l1: [new NamedNode('c'), new NamedNode('d'), new NamedNode('e')],
+        l2: [new Literal('c'), new Literal('d'), new Literal('e')],
+      };
+
+      const writer = new Writer({ lists });
+      writer.addQuad(new BlankNode('l1'), new NamedNode('b'), new BlankNode('l2'));
+      writer.addQuad(new NamedNode('a3'), new NamedNode('b'), new BlankNode('m3'));
+      writer.end((error, output) => {
+        output.should.equal('(<c> <d> <e>) <b> ("c" "d" "e").\n' +
+          '<a3> <b> _:m3.\n');
+        done(error);
+      });
+    });
+
+    it('should accept triples in bulk', done => {
+      const writer = new Writer();
       writer.addQuads([new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c')),
         new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('d'))]);
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('<a> <b> <c>, <d>.\n');
         done(error);
       });
     });
 
-    it('should not allow writing after end', function (done) {
-      var writer = N3Writer();
+    it('should not allow writing after end', done => {
+      const writer = new Writer();
       writer.addQuad(new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c')));
       writer.end();
-      writer.addQuad(new Quad(new NamedNode('d'), new NamedNode('e'), new NamedNode('f')), function (error) {
+      writer.addQuad(new Quad(new NamedNode('d'), new NamedNode('e'), new NamedNode('f')), error => {
         error.should.be.an.instanceof(Error);
         error.should.have.property('message', 'Cannot write because the writer has been closed.');
         done();
       });
     });
 
-    it('should write simple triples in N-Quads mode', function (done) {
-      var writer = N3Writer({ format: 'N-Quads' });
+    it('should write simple triples in N-Quads mode', done => {
+      const writer = new Writer({ format: 'N-Quads' });
       writer.addQuad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c'));
       writer.addQuad(new NamedNode('a'), new NamedNode('b'), new NamedNode('d'));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         output.should.equal('<a> <b> <c> .\n<a> <b> <d> .\n');
         done(error);
       });
     });
 
-    it('should write simple quads in N-Quads mode', function (done) {
-      var writer = N3Writer({ format: 'N-Quads' });
-      var called = false;
+    it('should write simple quads in N-Quads mode', done => {
+      const writer = new Writer({ format: 'N-Quads' });
+      let called = false;
       function callback() { called = true; }
       writer.addQuad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c'), callback);
       writer.addQuad(new NamedNode('a'), new NamedNode('b'), new NamedNode('d'), new NamedNode('g'));
-      writer.end(function (error, output) {
+      writer.end((error, output) => {
         called.should.be.true;
         output.should.equal('<a> <b> <c> .\n<a> <b> <d> <g> .\n');
         done(error);
       });
     });
 
-    it('should end when the end option is not set', function (done) {
-      var outputStream = new QuickStream(), writer = N3Writer(outputStream, {});
+    it('should end when the end option is not set', done => {
+      const outputStream = new QuickStream(), writer = new Writer(outputStream, {});
       outputStream.should.have.property('ended', false);
-      writer.end(function () {
+      writer.end(() => {
         outputStream.should.have.property('ended', true);
         done();
       });
     });
 
-    it('should end when the end option is set to true', function (done) {
-      var outputStream = new QuickStream(), writer = N3Writer(outputStream, { end: true });
+    it('should end when the end option is set to true', done => {
+      const outputStream = new QuickStream(), writer = new Writer(outputStream, { end: true });
       outputStream.should.have.property('ended', false);
-      writer.end(function () {
+      writer.end(() => {
         outputStream.should.have.property('ended', true);
         done();
       });
     });
 
-    it('should not end when the end option is set to false', function (done) {
-      var outputStream = new QuickStream(), writer = N3Writer(outputStream, { end: false });
+    it('should not end when the end option is set to false', done => {
+      const outputStream = new QuickStream(), writer = new Writer(outputStream, { end: false });
       outputStream.should.have.property('ended', false);
-      writer.end(function () {
+      writer.end(() => {
         outputStream.should.have.property('ended', false);
         done();
       });
     });
+
+    it('should serialize a triple with a triple with mixed component types as subject', () => {
+      const writer = new Writer();
+      writer.quadToString(new Quad(new BlankNode('b1'), new NamedNode('b'), new Literal('l1')), new NamedNode('b'), new NamedNode('c')).should.equal('<<_:b1 <b> "l">> <b> <c> .\n');
+    });
+
+    it('should serialize a triple with a triple with iris as subject', () => {
+      const writer = new Writer();
+      writer.quadToString(new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c')), new NamedNode('b'), new NamedNode('c')).should.equal('<<<a> <b> <c>>> <b> <c> .\n');
+    });
+
+    it('should serialize a triple with a triple with blanknodes as subject', () => {
+      const writer = new Writer();
+      writer.quadToString(new Quad(new BlankNode('b1'), new BlankNode('b2'), new BlankNode('b3')), new NamedNode('b'), new NamedNode('c')).should.equal('<<_:b1 _:b2 _:b3>> <b> <c> .\n');
+    });
+
+    it('should serialize a triple with a triple as object', () => {
+      const writer = new Writer();
+      writer.quadToString(new NamedNode('a'), new NamedNode('b'), new Quad(new BlankNode('b1'), new NamedNode('b'), new Literal('l1'))).should.equal('<a> <b> <<_:b1 <b> "l">> .\n');
+    });
+
+    it('should serialize a triple with a triple with iris as object', () => {
+      const writer = new Writer();
+      writer.quadToString(new NamedNode('a'), new NamedNode('b'), new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c'))).should.equal('<a> <b> <<<a> <b> <c>>> .\n');
+    });
+
+    it('should serialize a triple with a triple with blanknodes as object', () => {
+      const writer = new Writer();
+      writer.quadToString(new NamedNode('a'), new NamedNode('b'), new Quad(new BlankNode('b1'), new BlankNode('b2'), new BlankNode('b3'))).should.equal('<a> <b> <<_:b1 _:b2 _:b3>> .\n');
+    });
+
+    it('should serialize a quad with a triple with mixed component types as subject', () => {
+      const writer = new Writer();
+      writer.quadToString(new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c')), new NamedNode('b'), new NamedNode('c'), new NamedNode('g')).should.equal('<<<a> <b> <c>>> <b> <c> <g> .\n');
+    });
+
+    it('should serialize a quad with a triple as object', () => {
+      const writer = new Writer();
+      writer.quadToString(new NamedNode('a'), new NamedNode('b'), new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c')), new NamedNode('g')).should.equal('<a> <b> <<<a> <b> <c>>> <g> .\n');
+    });
+
+    it('should serialize a quad with a quad as subject', () => {
+      const writer = new Writer();
+      writer.quadToString(new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c'), new NamedNode('g')), new NamedNode('b'), new NamedNode('c'), new NamedNode('g')).should.equal('<<<a> <b> <c> <g>>> <b> <c> <g> .\n');
+    });
+
+    it('should serialize a quad with a quad as object', () => {
+      const writer = new Writer();
+      writer.quadToString(new NamedNode('a'), new NamedNode('b'), new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c'), new NamedNode('g')), new NamedNode('g')).should.equal('<a> <b> <<<a> <b> <c> <g>>> <g> .\n');
+    });
+
+    it('should serialize a triple with a quad as subject', () => {
+      const writer = new Writer();
+      writer.quadToString(new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c'), new NamedNode('g')), new NamedNode('b'), new NamedNode('c')).should.equal('<<<a> <b> <c> <g>>> <b> <c> .\n');
+    });
+
+    it('should serialize a triple with a quad as object', () => {
+      const writer = new Writer();
+      writer.quadToString(new NamedNode('a'), new NamedNode('b'), new Quad(new NamedNode('a'), new NamedNode('b'), new NamedNode('c'), new NamedNode('g'))).should.equal('<a> <b> <<<a> <b> <c> <g>>> .\n');
+    });
   });
 });
 
 function shouldSerialize(/* prefixes?, tripleArrays..., expectedResult */) {
-  var tripleArrays = Array.prototype.slice.call(arguments),
+  const tripleArrays = Array.prototype.slice.call(arguments),
       expectedResult = tripleArrays.pop(),
       prefixes = tripleArrays[0] instanceof Array ? null : tripleArrays.shift();
 
   return function (done) {
-    var outputStream = new QuickStream(),
-        writer = N3Writer(outputStream, prefixes);
+    const outputStream = new QuickStream(),
+        writer = new Writer(outputStream, prefixes);
     (function next() {
-      var item = tripleArrays.shift();
+      const item = tripleArrays.shift();
       if (item) {
-        var subject   = typeof item[0] === 'string' ? fromId(item[0]) : item[0];
-        var predicate = typeof item[1] === 'string' ? fromId(item[1]) : item[1];
-        var object    = typeof item[2] === 'string' ? fromId(item[2]) : item[2];
-        var graph     = typeof item[3] === 'string' ? fromId(item[3]) : item[3];
+        const subject   = typeof item[0] === 'string' ? termFromId(item[0]) : item[0];
+        const predicate = typeof item[1] === 'string' ? termFromId(item[1]) : item[1];
+        const object    = typeof item[2] === 'string' ? termFromId(item[2]) : item[2];
+        const graph     = typeof item[3] === 'string' ? termFromId(item[3]) : item[3];
         writer.addQuad(new Quad(subject, predicate, object, graph), next);
       }
       else
-        writer.end(function (error) {
+        writer.end(error => {
           try {
             outputStream.result.should.equal(expectedResult);
             outputStream.should.have.property('ended', true);
@@ -594,7 +768,8 @@ function shouldSerialize(/* prefixes?, tripleArrays..., expectedResult */) {
 }
 
 function QuickStream() {
-  var stream = { ended: false }, buffer = '';
+  const stream = { ended: false };
+  let buffer = '';
   stream.write = function (chunk, encoding, callback) {
     buffer += chunk;
     callback && callback();
